use std::{io::ErrorKind, fs, cell::RefCell, rc::Rc};

use log::{warn, debug, info};
use mlua::{Lua, ToLua, Function};
use tokio::sync::mpsc::Sender;
use std::sync::mpsc::Receiver;
use uuid::Uuid;

use self::{plugin::Plugin, event::{PluginEvent, ServerEvent}, server::{CellServer, Server}};

pub mod event;
mod plugin;
mod server;

#[derive(Clone, Copy, Eq, PartialEq)]
struct UuidUD(Uuid);

impl mlua::UserData for UuidUD {
	fn add_methods<'lua, M: mlua::UserDataMethods<'lua, Self>>(methods: &mut M) {
		methods.add_meta_method("__tostring",
			|lua, this, ()| this.0.to_string().to_lua(lua)
		);
		methods.add_meta_method("__eq",
			|lua, this, UuidUD(uuid)| (this.0 == uuid).to_lua(lua)
		);
	}
}

pub fn load_plugins(lua: &Lua) -> Vec<Plugin> {
	debug!("loading plugins");
	let entries = match fs::read_dir("./plugins") {
		Ok(n) => n,
		Err(e) if e.kind() == ErrorKind::NotFound => {
			warn!("plugins directory does not exist, creating");
			if let Err(e) = fs::create_dir("./plugins") {
				warn!("failed to create plugins directory: {e}");
			}
			return Vec::new()
		},
		Err(e) => {
			warn!("failed to load plugins: {e}");
			return Vec::new()
		}
	};
	let mut plugins = Vec::new();
	for entry in entries {
		let entry = match entry {
			Ok(e) => e,
			Err(e) => {
				warn!("error reading entry: {e}");
				continue
			}
		};
		let path = entry.path();
		let spath = path.to_str().unwrap_or("<non-unicode path>");
		debug!("loading plugin at {spath}");
		match Plugin::load(&entry, lua) {
			Ok(p) => {
				info!("loaded plugin {}", p.id);
				plugins.push(p);
			}
			Err(e) => warn!("error loading plugin at {spath}: {e}"),
		}
	}
	plugins
}

pub async fn run_plugins(tx: Sender<PluginEvent>, rx: Receiver<ServerEvent>) {
	let lua = unsafe {
		mlua::Lua::unsafe_new_with(
			mlua::StdLib::ALL_SAFE | mlua::StdLib::DEBUG,
			mlua::LuaOptions::default()
		)
	};

	let plugins = load_plugins(&lua);

	let server = Rc::new(RefCell::new(Server {
		tx,
		names: Default::default(),
		info: Default::default(),
	}));

	for plugin in &plugins {
		if let Ok(init) = plugin.table.get::<_, Function>("init") {
			if let Err(e) = init.call::<_, ()>((plugin.table.clone(), )) {
				warn!("error initializing plugin {}: {e}", plugin.id);
			}
		}
	}

	while let Ok(ev) = rx.recv() {
		match ev {
			ServerEvent::Join(player, info) => {
				{
					let mut guard = server.borrow_mut();
					guard.names.insert(player.uuid, player.name.clone());
					guard.info.insert(player.uuid, info);
				}
				for plugin in &plugins {
					if let Ok(f) = plugin.table.get::<_, Function>("on_join") {
						let res = f.call::<_, ()>((
							plugin.table.clone(),
							CellServer(server.clone()),
							UuidUD(player.uuid),
							player.name.clone()
						));
						if let Err(e) = res {
							warn!("error responding to join event in plugin {}: {e}", plugin.id);
						}
					}
				}
			},
			ServerEvent::Leave(player) => {
				for plugin in &plugins {
					if let Ok(f) = plugin.table.get::<_, Function>("on_leave") {
						let res = f.call::<_, ()>((
							plugin.table.clone(),
							CellServer(server.clone()),
							UuidUD(player.uuid),
							player.name.clone()
						));
						if let Err(e) = res {
							warn!("error responding to leave event in plugin {}: {e}", plugin.id);
						}
					}
				}
				{
					let mut guard = server.borrow_mut();
					guard.names.remove(&player.uuid);
					guard.info.remove(&player.uuid);
				}
			}
		}
	}
}
