use std::{collections::HashMap, cell::RefCell, rc::Rc};

use mlua::{Value, ToLua};
use tokio::sync::mpsc::Sender;
use uuid::Uuid;

use crate::ClientInfo;

use super::{event::PluginEvent, UuidUD};

impl<'lua> ToLua<'lua> for ClientInfo {
    fn to_lua(self, lua: &'lua mlua::Lua) -> mlua::Result<Value<'lua>> {
        let table = lua.create_table()?;
		table.set("addr", self.addr)?;
		table.set("port", self.port)?;
		table.set("proto_name", self.proto_name)?;
		table.set("proto_version", self.proto_version)?;
		Ok(Value::Table(table))
    }
}

pub struct Server {
	pub tx: Sender<PluginEvent>,
	pub names: HashMap<Uuid, String>,
	pub info: HashMap<Uuid, ClientInfo>,
}

#[derive(Clone)]
pub struct CellServer(pub Rc<RefCell<Server>>);

impl mlua::UserData for CellServer {
	fn add_methods<'lua, M: mlua::UserDataMethods<'lua, Self>>(methods: &mut M) {
		methods.add_method("players", |lua, this, ()| {
			let table = lua.create_table()?;
			for (k, v) in &this.0.borrow().names {
				table.set(UuidUD(*k), v.clone())?;
			}
			Ok(table)
		});
		methods.add_method("get_name", |lua, this, UuidUD(uuid)| {
			match this.0.borrow().names.get(&uuid) {
				Some(s) => s.clone().to_lua(lua),
				None => Ok(Value::Nil),
			}
		});
		methods.add_method("get_info", |lua, this, UuidUD(uuid)| {
			match this.0.borrow().info.get(&uuid) {
				Some(i) => i.clone().to_lua(lua),
				None => Ok(Value::Nil)
			}
		});
		methods.add_async_method("kick", |_, CellServer(this), (UuidUD(uuid), msg)| async move {
			let tx = this.borrow().tx.clone();
			tx
				.send(PluginEvent::Kick(uuid, msg))
				.await
				.map_err(|e| mlua::Error::RuntimeError(e.to_string()))
		});
	}
}

