use std::io::{Read, Write};

use uuid::Uuid;

use crate::{JsonValue, ser::Position};

mod common;
mod v1_19_4;
mod v1_20_1;

pub use common::COMMON as COMMON;

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub enum ProtocolState {
	Handshake = 0,
	Status = 1,
	Login = 2,
	Play = 3,
}

impl TryFrom<i32> for ProtocolState {
    type Error = ();
    fn try_from(value: i32) -> std::result::Result<Self, Self::Error> {
		match value {
			0 => Ok(Self::Handshake),
			1 => Ok(Self::Status),
			2 => Ok(Self::Login),
			_ => Err(())
		}
    }
}

#[derive(Clone, Debug)]
pub enum ClientPacket {
	Handshake {
		version: i32,
		addr: String,
		port: u16,
		next_state: i32
	},
	StatusRequest,
	PingRequest(i64),
	LoginStart {
		name: String,
		uuid: Option<Uuid>
	},
	PluginMessage {
		channel: String,
		data: Vec<u8>
	},
	KeepAlive(i64),
}

#[derive(Clone, Debug)]
pub enum ServerPacket {
	StatusResponse(JsonValue),
	PingResponse(i64),
	LoginDisconnect(JsonValue),
	LoginSuccess {
		name: String,
		uuid: Uuid
	},
	JoinGame {
		eid: i32,
		gamemode: u8,
		hardcode: bool,
	},
	PlayDisconnect(JsonValue),
	KeepAlive(i64),
	PluginMessage {
		channel: String,
		data: Vec<u8>
	},
	ChunkData {
		x: i32,
		z: i32,
	},
	SetDefaultSpawn { pos: Position, angle: f32 },
}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub struct Protocol {
	pub name: &'static str,
	pub version: i32,
	pub encode: fn(w: Box<&mut dyn Write>, state: ProtocolState, ev: ServerPacket) -> anyhow::Result<()>,
	pub decode: fn(r: Box<&mut dyn Read>, state: ProtocolState, len: i32, id: i32) -> anyhow::Result<Option<ClientPacket>>,
}

pub const fn get_protocol(version: i32) -> Option<Protocol> {
	match version {
		v1_20_1::VERSION => Some(v1_20_1::PROTOCOL),
		v1_19_4::VERSION => Some(v1_19_4::PROTOCOL),
		_ => None,
	}
}

