use std::{collections::HashMap, time::Duration, net::ToSocketAddrs};

use log::{info, error};
use tokio::{net::TcpListener, sync::mpsc::{Sender, self}, select};

use crate::{client::run_client, Player, ClientInfo, client::event::{ServerEvent, ClientEvent}, plugin::{run_plugins, event::PluginEvent}};

type PServerEvent = crate::plugin::event::ServerEvent;

pub struct Client {
	tx: Sender<ServerEvent>,
	keepalive: Option<i64>,
	info: Option<ClientInfo>,
	player: Option<Player>,
}

pub async fn run_server() -> Result<(), Box<dyn std::error::Error>> {
	env_logger::init();

	let (ps_tx, ps_rx) = std::sync::mpsc::channel();
	let (pc_tx, mut pc_rx) = mpsc::channel(256);

	tokio::spawn(run_plugins(pc_tx, ps_rx));

	let socket_addr = "0.0.0.0:25565".to_socket_addrs()?.next().expect("invalid server address");
	let listener = TcpListener::bind(socket_addr).await?;

	let mut next_id = 0i32;
	let (gc_tx, mut gc_rx) = mpsc::channel(16);
	let mut keepalive = tokio::time::interval(Duration::from_secs(15));
	let mut clients = HashMap::new();

	info!("listening on {socket_addr}");

	loop { select! {
		conn = listener.accept() => {
			let gc_tx = gc_tx.clone();
			let (stream, addr) = conn?;

            let id = next_id;
			next_id = next_id.wrapping_add(1);

            let (gs_tx, gs_rx) = mpsc::channel(16);
            clients.insert(id, Client {
                tx: gs_tx,
				keepalive: None,
				info: None,
                player: None,
            });
			info!("#{id} connected from {addr}");
            tokio::spawn(async move {
				let c_tx2 = gc_tx.clone();
				match run_client(id, stream, gc_tx, gs_rx).await {
					Ok(_) => info!("client {id} disconnected"),
					Err(e) => error!("client {id} error: {e}"),
				}
				c_tx2.send((id, ClientEvent::Disconnect)).await.unwrap();
            });
		},
		_ = keepalive.tick() => {
			for client in clients.values_mut() {
				if client.keepalive.is_some() {
					client.tx.send(ServerEvent::Disconnect(Some("Failed to respond to keep alives".to_owned()))).await?;
				} else if client.player.is_some() {
					let data = 1;
					client.keepalive = Some(data);
					client.tx.send(ServerEvent::KeepAlive(data)).await?;
				}
			}
		},
		ev = pc_rx.recv() => {
			let Some(ev) = ev else {
				return Err("plugin channel closed".into());
			};
			match ev {
				PluginEvent::Kick(u, msg) => {
					for client in clients.values() {
						if let Some(Player { uuid, .. }) = client.player {
							if uuid == u {
								client.tx.send(ServerEvent::Disconnect(msg.clone())).await?;
								break;
							}
						}
					}
				}
			}
		},
		ev = gc_rx.recv() => {
			let Some(ev) = ev else {
				return Err("reciever closed".into());
			};
			let id = ev.0;
			let client = clients.get_mut(&id).unwrap();
			match ev.1 {
				ClientEvent::Handshake(info) => client.info = Some(info),
				ClientEvent::Join(player) => {
					client.player = Some(player.clone());
					ps_tx.send(PServerEvent::Join(player, client.info.clone().unwrap()))?;
				},
				ClientEvent::Disconnect => {
					let player = clients.get(&id).unwrap().player.clone();
					clients.remove(&id);
					if let Some(player) = player {
						ps_tx.send(PServerEvent::Leave(player))?;
					}
				},
				ClientEvent::KeepAlive(data) => {
					if client.keepalive == Some(data) {
						client.keepalive = None;
					} else {
						client.keepalive = Some(0);
					}
				}
			}
		}
	} }
}
